.selector-0
  .selector-1
    color: rebeccapurple
  /*comments*/
    /*other comment*/

.selector-0
  .selector-1
    color: rebeccapurple
    /*comments*/
    /*other comment*/

.selector-0
  .selector-1
    color: rebeccapurple
  /*comments*/
  /*other comment*/
