::selection {
  color: rebeccapurple; }
::-webkit-selection {
  color: rebeccapurple; }
::-webkit-selection, ::selection {
  color: rebeccapurple; }
: -webkit-selection, : selection {
  color: rebeccapurple; }
