@font-face
  font-family: "test1 \" \' ' */ // */ " // valid comment
  src: url("http://fonts/test1.eot?#iefix") format("embedded-opentype")
@font-face
  font-family: 'test2 \' \" " */ // */ ' // valid comment
  src: url('http://fonts/test2.eot?#iefix') format('embedded-opentype')
@font-face
  font-family: test3 // valid comment
  src: url(http://fonts/test3.eot?#iefix) format('embedded-opentype')
