.error {
  border: 1px #f00;
  background-color: #fdd; }

.seriousError {
  @extend .error;
  border-width: 3px; }

.criticalError {
  @extend .seriousError;
  position: fixed;
  top: 10%;
  bottom: 10%;
  left: 10%;
  right: 10%; }
