/* not closed comment
  not closed comment
 not closed comment */
outer { /* comment // ignore */
  inner /* ignore // comment */ {
    color: red;
    /* before closing */ }
  /* on parent */ }

/* some before */
/* not closed comment
   //   some more lines
 edge case (still works
   but one char is overwritten */

/* this
  is also

 	a comment! */
