/* not closed comment
  not closed comment
 not closed comment */
outer {
  inner {
    color: red; } }

/* not closed comment
   //   some more lines
 edge case (still works
   but one char is overwritten */
