@charset "UTF-8";
[data-icon=test-1]:before {
  content: "\\";
}

[data-icon=test-2]:before {
  content: "'";
}

[data-icon=test-3]:before {
  content: '"';
}

[data-icon=test-4]:before {
  content: "\\";
}

[data-icon=test-5]:before {
  content: "'";
}

[data-icon=test-6]:before {
  content: '"';
}

.émotion {
  blah: hi;
}

.édition {
  blah: hi;
}

.édition {
  blah: hi;
}
