// not closed comment
//not closed comment
//ot closed comment
outer {
  inner {
    color: red; } }

// not closed comment
// //   some more lines
//dge case (still works
// but one char is overwritten
